##############################################
#	Made by Atle Dreier
#	Force Feedback Clipping Meter
#
##############################################

# Rewrite v4.3.3
# Fixed "Gain stuck at 200" bug



#Libraries
import ac
import acsys
import os
import sys
import configparser
import traceback
import csv

#Constant definitions
Version="4.3.3"

K=0.12 #Lowpass filter constant for the FFB signal

DMode=1
Cutoff=0.03

#Variable initialization
dT=0
dTR=0
OldForce=0
UIClipIndicator=1
AvgForce=0.01
ffbMultiplier=1
AvgForce=0.01
GainChange=1
DynamicTarget=1
OldFFBMultiplier=1
CarGain=1
OldGain=1
ComboGain=0
Histogram = []
DynamicMode=0
OldRAW=0

#Main app 
def acMain(acVersion):

#Global variables
	global UIPreferredProfile, CurrentProfile, ProfileButton #for the profile change button
	global AutoMode,ModeButton #for the Mode change button
	global LogButton,CarID,TrackID,TrackConfigID #for the logging button
	global graph,TargetGain,appWindow,UIClipIndicator,mildClipLabel,hardClipLabel,filterLabel,MainGain,CarGain,CarGainLabel #for the renderloop
	global config,Combo,ComboPath,messageLabel, UserFF,userffPath,ComboGain,Version
	global DMode,Histogram,Cutoff,DispChangeButton,DynamicModeButton

#Intializing configparser
	config = configparser.ConfigParser()
	Combo = configparser.ConfigParser()
	UserFF = configparser.ConfigParser()
	
#IMPORTANT : Path to change if the application displays "Controls.ini is missing. Using v1.0"
	#settingsPath = os.path.expanduser("~\Documents\Assetto Corsa\cfg\controls.ini")
	userffPath = os.path.expanduser(r"~\Documents\Assetto Corsa\cfg\user_ff.ini")
	#settingsPath = os.path.abspath("C:\\Users\Atle\Documents\Assetto Corsa\cfg\controls.ini"")
	ComboPath= "apps/python/FFBClip/Config/Combos.ini"
	
	ac.log("Paths set")

#Get the car name and write it to console
	CarID = ac.getCarName(0)
	TrackID = ac.getTrackName(0)
	TrackConfigID = ac.getTrackConfiguration(0)
	ac.console(CarID)

#Get the last saved gain for the user's car if it exist, otherwise set 100% by default
#Cap the value to 200%
#Write the gain to console and set auto or manual mode


	
#Write the AC settings path to console. Pure debug line
	ac.console(os.getcwd())
	
#Read the ini file and set appropriate values
	if os.path.exists("apps/python/FFBClip/Config/FFBClip.ini"):
		ac.console("FFBClip.ini found")
		config.read("apps/python/FFBClip/Config/FFBClip.ini")
		if config.has_section("TargetGain"):
			ac.console("TargetGain read")
			TargetGain = float(config.get("TargetGain","TargetFFB"))
			ac.console(str(TargetGain))
		else:
			ac.console("Section not found")
			TargetGain=1.0
		
		if config.has_section("Minimized"):
			ac.console("Minimized read")
			UIMinimized = float(config.get("Minimized","RunMinimized"))
			ac.console(str(UIMinimized))
		else:
			ac.console("Section not found")
			UIMinimized=0
			
		if config.has_section("IndicateClipping"):
			ac.console("ClipIndi read")
			UIClipIndicator = float(config.get("IndicateClipping","IndicateClipping"))
			ac.console(str(UIClipIndicator))
		else:
			ac.console("Section not found")
			
		if config.has_section("PreferredProfile"):
			ac.console("PreferredProfile read")
			UIPreferredProfile = float(config.get("PreferredProfile","PreferredProfile"))
			CurrentProfile=UIPreferredProfile
			ac.console(str(UIPreferredProfile))
		else:
			ac.console("Section not found")
			
		if config.has_section("PreferredMode"):
			ac.console("PreferAuto read")
			UIPreferAuto = float(config.get("PreferredMode","PreferAutoMode"))
			ac.console("PreferAuto:" + str(UIPreferAuto))
			if UIPreferAuto==1:
				AutoMode=1
				errorMessage = "Forced auto mode"
		else:
			ac.console("Section not found")
			
		if config.has_section("PreferredMode"):
			ac.console("PreferManual read")
			UIPreferManual = float(config.get("PreferredMode","PreferManualMode"))
			ac.console("Prefermanual:"+str(UIPreferManual))
			if UIPreferManual==1:
				AutoMode=0
				errorMessage = "Forced manual mode"
		else:
			ac.console("Section not found")
			
		if config.has_section("PreferredMode"):
			ac.console("PreferGlobal read")
			UIPreferGlobal = float(config.get("PreferredMode","PreferGlobalMode"))
			ac.console("PreferGlobal:"+str(UIPreferGlobal))
			if UIPreferGlobal==1:
				AutoMode=0
				errorMessage = "Global mode"
		else:
			ac.console("Section not found")
			
		if config.has_section("PreferredMode"):
			ac.console("PreferGraph read")
			UIPreferGraph = float(config.get("PreferredMode","PreferGraph"))
			ac.console("PreferGraph:"+str(UIPreferGlobal))
			if UIPreferGraph==1:
				DMode=0
		else:
			ac.console("Section not found")
			
		ac.log("Controls.ini section start")
		
		if config.has_section("Controls_ini_Path"):
			ac.console("Path Read")
			settingsPath=config.get("Controls_ini_Path","Path")
			if settingsPath !="":
				ac.console(settingsPath)
				#errorMessage = "Settingspath found in .ini"
			else:
				settingsPath = os.path.expanduser("~\Documents\Assetto Corsa\cfg\controls.ini")
				ac.console(settingsPath)
				ac.log(settingsPath)
		else:
			ac.console("Section not found")
			ac.log("Controls.ini section not found")
			
#If the ini file isn't found we set default values		
	else:
		ac.console("FFBClip.ini not found")
		ac.log("FFBClip.ini not found")
		errorMessage = "FFBClip.ini not found"
		TargetGain=1
		UIClipIndicator=1
		UIMinimized=0
		UIPreferredProfile=0
		UIPreferAuto=1
	
	ac.log("FFBClip.ini section end")	
	
	ac.log("Combos.ini section start")
	if os.path.exists(ComboPath):
		ac.console("Combos.ini found")
		Combo.read(ComboPath)
		ac.log("loaded")
		if Combo.has_option(TrackID+"_"+TrackConfigID,CarID):
			ac.log("Has section and option")
			CarGain = float(Combo.get(TrackID+"_"+TrackConfigID,CarID))
			ComboGain=1
			AutoMode=0
			errorMessage = "Old combo, manual mode"
			if CarGain>2:
				CarGain=2
			ac.console("Combos cargain:"+str(CarGain))
		else:
			ac.console("Combos.ini or section not found")
			CarGain = 1.0
			AutoMode=1
			ComboGain=0
			errorMessage = "New combo, Auto mode"
	else:
		CarGain=1.0
		AutoMode=1
		errorMessage="Combos.ini created"
		with open('apps/python/FFBClip/Config/Combos.ini', 'w') as configfile:
			Combo.write(configfile)
		
	ac.log("Combo loaded")
	
	#if UIPreferGlobal==1:
	ac.log("UserFF section start")
	ac.log("Cargain before USerFF:" + str(CarGain))
	if os.path.exists(userffPath)and UIPreferGlobal==1:
		ac.console("user_ff.ini found")
		UserFF.read(userffPath)
		if UserFF.has_section(CarID):
			CarGain = float(UserFF.get(CarID,"VALUE"))
			if CarGain>2:
				CarGain=2
			AutoMode=0
			ac.console(str(CarGain))
			errorMessage = "Global car gain used"
		else:
			ac.console("user_ff.ini not found")
			CarGain = 1.0
			AutoMode=1
			errorMessage = "New car, global mode"
	ac.log("UserFF section end")
	ac.log("Cargain after UserFF:"+str(CarGain))
			
#Get the main menu gain from Controls.ini. If that doesn't exist or isn't found we use 100%
	if os.path.exists(settingsPath):
		ac.log("Controls.ini found")
		config.read(settingsPath)
		MainGain = float(config.get("STEER","FF_GAIN"))
		ac.log("MainGain:"+str(MainGain))
	else:
	#	#exists(path) is false then we use the v1.0 value and display it
		ac.log("MainGain not found")
		MainGain = 1.0
		errorMessage = "Controls.ini is missing. Using 100%"		
	
	if UIPreferAuto==1:
		AutoMode=1
		errorMessage = "Forced auto mode"
	
	if UIPreferManual==1:
		AutoMode=0
		errorMessage = "Forced manual mode"
		
	if UIPreferGlobal==1:
		AutoMode=0
		errorMessage = "Forced global mode"
		
	CarGainNow=round(ac.getCarFFB()/100,2)
	GainChange=CarGain-CarGainNow
	ac.setCarFFB(GainChange*100)
	
	ac.log("Setting up UI")
#Set up the main window
	if UIMinimized == 0:
		ac.console("Maximized")
		appWindow = ac.newApp("FFBClip")
		ac.setSize(appWindow,314,260)
	else:
		ac.console("Minimized")
		appWindow = ac.newApp("FFBClip")
		ac.setSize(appWindow,314,120)
	
	ac.setTitle(appWindow,"FFBClip v"+Version)
#Set up all the UI elements
	
#Set up the graph if maximized
	if UIMinimized == 0:
		graph = ac.addGraph(appWindow,"FFB_Graph")
		ac.setPosition(graph,7,103)
		ac.setSize(graph,300,150)
		ac.setRange(graph,-0.05,2.1,300)
		ac.addSerieToGraph(graph,0,1,0)
		ac.addSerieToGraph(graph,1,0,0)
		ac.addSerieToGraph(graph,1,1,0)
		ac.addSerieToGraph(graph,1,1,1)
		ac.addSerieToGraph(graph,0,0,0)
		
#Set up the error message
	messageLabel = ac.addLabel(appWindow,errorMessage)
	ac.setPosition(messageLabel,7,60)
#The what to set gain to label
	filterLabel = ac.addLabel(appWindow,CarID)
	ac.setPosition(filterLabel,7,30)
	ac.setFontColor(filterLabel, 1, 1, 0, 1)	
#The current gain label
	CarGainLabel = ac.addLabel(appWindow,"Current gain:")
	ac.setPosition(CarGainLabel,7,45)

#Active UI elements

#DynamicModeButton
	DynamicModeButton = ac.addButton(appWindow,"Static")
	ac.setSize(DynamicModeButton,72,20)
	ac.addOnClickedListener(DynamicModeButton, DynamicChange)
	ac.setPosition(DynamicModeButton,157,80)
	
#The profile selector button
	ProfileButton = ac.addButton(appWindow,"Normal")
	ac.setPosition(ProfileButton,237,57)
	ac.setSize(ProfileButton,70,20)
	ac.addOnClickedListener(ProfileButton, ProfileChange)
	if CurrentProfile == 0:
		ac.setText(ProfileButton,"Normal")
	if CurrentProfile == 1:
		ac.setText(ProfileButton,"Strong")
	if CurrentProfile == 2:
		ac.setText(ProfileButton,"Maximal")
	if CurrentProfile == 3:
		ac.setText(ProfileButton,"Light")
	if CurrentProfile == 4:
		ac.setText(DynamicModeButton,"Dynamic")
		DynamicMode=1
		AutoMode=1

#The mode selector button
	if AutoMode==0:
		ModeButton = ac.addButton(appWindow,"Manual")
	else:
		ModeButton = ac.addButton(appWindow,"Auto")
	ac.setSize(ModeButton,70,20)
	ac.addOnClickedListener(ModeButton, ModeChange)
	ac.setPosition(ModeButton,237,80)

#LogButton
	LogButton = ac.addButton(appWindow,"Reset")
	ac.setSize(LogButton,72,20)
	ac.addOnClickedListener(LogButton, LogChange)
	ac.setPosition(LogButton,82,80)
	
#DisplayModeButton
	DispChangeButton = ac.addButton(appWindow,"Histogram")
	if DMode==0:
		ac.setText(DispChangeButton,"Graph")
	ac.setSize(DispChangeButton,72,20)
	ac.addOnClickedListener(DispChangeButton, DispChange)
	ac.setPosition(DispChangeButton,7,80)

#initialize the Histogram
	for n in range(0,300):
		Histogram.append(0)
		

#Call the main render loop
	ac.addRenderCallback(appWindow,onFormRender)
	ac.log("Just before render starts")
	ac.log("ComboGain:"+str(ComboGain))
	return "FFBClip"

def acUpdate(deltaT):

	global graph,dT,K,OldForce,appWindow,UIClipIndicator,hardClip,mildClip,hardClip,mildClip,filterLabel,Logging
	global peak,MaxForce,Peak2,Peak3,PeakCount,peakforce,rise,OldFFBMultiplier,CurrentProfile,messageLabel,ModeButton
	global MainGain,CarGain,TargetGain,ffbMultiplier,AvgForce
	global AutoMode,GainChange,CarGainLabel,DynamicTarget,OldGain
	global CarID,TrackID,TrackConfigID,LastLap,LogFile,ComboGain
	global DMode,Histogram,Cutoff,CurrentForce,FilteredForce,OldRAW
	
#check if it's time to redraw
	dT += deltaT
	
#Read this frame's FFB and add it to the histogram
	CurrentForce = abs(ac.getCarState(0,acsys.CS.LastFF))

	n=int(CurrentForce*200)
	if n>0 and OldRAW != CurrentForce:
		a=Histogram.pop(n)
		if n<50:
			inc=n/50
		else:
			inc=1	
		a += inc
		Histogram.insert(n,a)
	
	OldRAW=CurrentForce
#Run the loop every 30ms
	if dT>0.03:
		dT=0 #Reset the loop timer
		
#Read the relevant values from the game
		#CurrentForce = abs(ac.getCarState(0,acsys.CS.LastFF))
	
#Find the new filtered force
		a=deltaT/(K+deltaT)
		FilteredForce = (OldForce*(1-a))+(CurrentForce*a)
		OldForce = FilteredForce

#Find gain based on histogram				
			
		HMax = max(Histogram)+1
		for i in range(299,1,-1):
			if (Histogram[i]/HMax) > Cutoff:
				ffbMultiplier=TargetGain/(i/200)
				#ac.log("FFBMult:"+str(ffbMultiplier)+" Cutoff: "+str(Cutoff)+" i: "+str(i)+" HMax: "+str(HMax)+" Hist/HMax: "+str(Histogram[i]/HMax))
				break
				
#Dynamic mode	
		if DynamicMode==1:
			if CurrentForce>(TargetGain-0.05):
				ffbMultiplier=TargetGain*(1/(CurrentForce))
			else:
				DynamicTarget=(1/MainGain)*TargetGain*(1+(TargetGain-CurrentForce))
				ffbMultiplier=((0.004*DynamicTarget)+(0.996*OldFFBMultiplier))
			OldFFBMultiplier=ffbMultiplier
				
#Value capping
		if DynamicTarget>2:
			DynamicTarget=2
		if ffbMultiplier > 2:
			ffbMultiplier=2
			CarGain=2
	


#Write the new gain to game and scale the peaks
		CarGain=round(ac.getCarFFB()/100,2)
			
		ffbMultiplier=round(ffbMultiplier,2)
		if AutoMode==1:
			GainChange=ffbMultiplier-CarGain	
		else:
			GainChange=OldGain-CarGain
			Change=CarGain/OldGain
			OldGain=CarGain
				
		#if AutoMode==0:
		#	if ffbMultiplier<(CarGain-0.15):
		#		ac.setText(filterLabel,"Gain is high. Should be : {:.0%}".format(ffbMultiplier))
		#	if ffbMultiplier>(CarGain+0.15):
		#		ac.setText(filterLabel,"Gain is low. Should be : {:.0%}".format(ffbMultiplier))

		#if AutoMode==1 and GainChange != 0:
		if GainChange != 0:# and AutoMode==1:
			#GainChange=(ffbMultiplier-((ac.getCarFFB()/100)))*100
			#Change=ffbMultiplier/CarGain
			if AutoMode==1:
				Change=ffbMultiplier/CarGain
				ac.setCarFFB(GainChange*100)
						

#Clipping indicator
		if UIClipIndicator == 1:
			if FilteredForce>(TargetGain-0.05):
				ClipIndi=(FilteredForce-(TargetGain-0.05))*20
				if ClipIndi>1:
					ClipIndi=1
			else:
				ClipIndi=0
			ac.setBackgroundColor(appWindow,ClipIndi,0.2,0.2)

def onFormRender(deltaT):

	global graph,dT,K,OldForce,appWindow,UIClipIndicator,samplesCount,hardClip,mildClip,hardClip,mildClip,filterLabel,Logging
	global peak,MaxForce,Peak2,Peak3,PeakCount,peakforce,rise,OldFFBMultiplier,CurrentProfile,messageLabel,ModeButton
	global MainGain,CarGain,TargetGain,ffbMultiplier,AvgForce
	global AutoMode,GainChange,CarGainLabel,DynamicTarget,OldGain
	global CarID,TrackID,TrackConfigID,LastLap,LogFile,ComboGain
	global DMode,Histogram,Cutoff,CurrentForce,FilteredForce,dTR
	
	dTR += deltaT	
	ac.setText(CarGainLabel,"Current gain : {:.0%}".format(CarGain))
	ac.setText(filterLabel,"FFB should be set to : {:.0%}".format(ffbMultiplier))
	if AutoMode==0:
		if ffbMultiplier<(CarGain-0.15):
			ac.setText(filterLabel,"Gain is high. Should be : {:.0%}".format(ffbMultiplier))
		if ffbMultiplier>(CarGain+0.15):
			ac.setText(filterLabel,"Gain is low. Should be : {:.0%}".format(ffbMultiplier))
		

	
	if DMode == 0:
		if dTR>0.03:
			dTR=0
			ac.addValueToGraph(graph,0,CurrentForce) #Green
			ac.addValueToGraph(graph,1,TargetGain) #Red
			ac.addValueToGraph(graph,2,FilteredForce) #Yellow
			#ac.addValueToGraph(graph,3,MaxForce) #debug WHITE
			#ac.addValueToGraph(graph,4,ffbMultiplier) #debug BLACK
	else:
		if dTR>1.0:
			dTR=0
			HMax = max(Histogram)+1
			for i in range(0,300):
				ac.addValueToGraph(graph,0,(Histogram[i]/HMax))
				
				#ac.addValueToGraph(graph,3,Cutoff)
				#if i>((1/ffbMultiplier)*200):
				#	ac.addValueToGraph(graph,1,1)
				#else:
				#	ac.addValueToGraph(graph,1,0)
					
				if i<200:
					ac.addValueToGraph(graph,2,0)
				else:
					ac.addValueToGraph(graph,2,1)
			
def acShutdown(*args):
	global TrackID,TrackConfigID,CarID,CarGain,config,Combo,ComboPath
	
	ac.log("shutdown")
	
	if os.path.exists(ComboPath):
		if Combo.has_section(TrackID+"_"+TrackConfigID):
			ac.log("found track")
			if Combo.has_option(TrackID+"_"+TrackConfigID,CarID):
				ac.log("Car found")
				Combo.remove_option(TrackID+"_"+TrackConfigID,CarID)
				Combo.set(TrackID+"_"+TrackConfigID,CarID,str(CarGain))
			else:
				ac.log("Car not found")
				Combo.set(TrackID+"_"+TrackConfigID,CarID,str(CarGain))
				ac.log("add")
		else:
			ac.log("Track not found")
			Combo.add_section(TrackID+"_"+TrackConfigID)
			
			Combo.set(TrackID+"_"+TrackConfigID,CarID,str(CarGain))
	else:
		ac.log("Not Found")
	
	with open('apps/python/FFBClip/Config/Combos.ini', 'w') as configfile:
		ac.log("Write")
		Combo.write(configfile)
	#Combo.close()
	ac.log("close")

def DispChange(val1,val2):
	global DMode,DispChangeButton,messageLabel
	
	if DMode==0:
		DMode=1
		ac.setText(DispChangeButton,"Histogram")
		ac.setText(messageLabel,"Display mode: Histogram")
	else:
		DMode=0
		ac.setText(DispChangeButton,"Graph")
		ac.setText(messageLabel,"Display mode: Graph")
	
def ModeChange(val1,val2):
	global AutoMode,ModeButton,messageLabel
	if AutoMode==0:
		AutoMode=1
		ac.setText(ModeButton,"Auto")
		ac.setText(messageLabel,"Auto mode engaged")
	else:
		AutoMode=0
		ac.setText(ModeButton,"Manual")
		ac.setText(messageLabel,"Manual mode engaged")
		
def ProfileChange(val1,val2):
	global UIPreferredProfile, CurrentProfile, ProfileButton,Cutoff,messageLabel
	CurrentProfile += 1
	if CurrentProfile > 3:
		CurrentProfile=0
	if CurrentProfile == 0:
		ac.setText(ProfileButton,"Normal")
		ac.setText(messageLabel,"Profile: Normal")
		Cutoff=0.03
	if CurrentProfile == 1:
		ac.setText(ProfileButton,"Strong")
		ac.setText(messageLabel,"Profile: Strong")
		Cutoff=0.07
	if CurrentProfile == 2:
		ac.setText(ProfileButton,"Maximal")
		ac.setText(messageLabel,"Profile: Maximal")
		Cutoff=0.15
	if CurrentProfile == 3:
		ac.setText(ProfileButton,"Light")
		ac.setText(messageLabel,"Profile: Light")
		Cutoff=0.0
		
	
def LogChange(val1,val2):
	global LogButton,Histogram,ffbMultiplier,AutoMode,messageLabel

	del Histogram[:]
		
	for n in range(0,300):
		Histogram.append(0)
	
	AutoMode=0
	ffbMultiplier=1.0
	CarGain=round(ac.getCarFFB()/100,2)	
	ffbMultiplier=round(ffbMultiplier,2)
	GainChange=ffbMultiplier-CarGain	
	if GainChange != 0:# and AutoMode==1:
		ac.setCarFFB(GainChange*100)
	
	ac.setText(messageLabel,"Reset solution")
def DynamicChange(val1, val2):
	global DynamicModeButton, DynamicMode, AutoMode,ModeButton,messageLabel
	
	DynamicMode += 1
	
	if DynamicMode > 1:
		DynamicMode=0
		ac.setText(DynamicModeButton,"Static")
		ac.setText(ModeButton,"Manual")
		ac.setText(messageLabel,"Dynamic mode disengaged")
		AutoMode=0
		                      
	if DynamicMode == 1:
		ac.setText(DynamicModeButton,"Dynamic")
		ac.setText(ModeButton,"Auto")
		ac.setText(messageLabel,"Dynamic mode engaged")
		AutoMode=1
		